"""
validate the use of shell processing for cmd.run on the salt command line
and in templating
"""

import os
import textwrap

import pytest

import salt.utils.files
import salt.utils.platform

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_on_freebsd,
]


@pytest.fixture
def test_cmd():
    if salt.utils.platform.is_windows():
        return 'where cmd whoami xcopy | find /c /v "" && echo saltines && echo duh > NUL 2>&1'
    else:
        return (
            "printf '%s\\n' first second third | wc -l ; "
            "export SALTY_VARIABLE='saltines' && echo $SALTY_VARIABLE ; "
            "echo duh &> /dev/null"
        )


@pytest.fixture
def ret_disabled():
    if salt.utils.platform.is_windows():
        return 'ERROR: Value for default option cannot be empty.\r\nType "WHERE /?" for usage.'
    else:
        return (
            "first\nsecond\nthird\n|\nwc\n-l\n;\nexport\nSALTY_VARIABLE=saltines"
            "\n&&\necho\n$SALTY_VARIABLE\n;\necho\nduh\n&>\n/dev/null"
        )


@pytest.fixture
def ret_enabled(cmd):
    if salt.utils.platform.is_windows():
        cmd_text = 'where cmd whoami xcopy | find /c /v ""'
        num_lines = cmd.run(cmd_text, python_shell=True)
        return f"{num_lines}{os.linesep}saltines"
    else:
        return f"3{os.linesep}saltines"


@pytest.fixture(scope="module")
def cmd(modules):
    return modules.cmd


@pytest.fixture(scope="module")
def state(modules):
    return modules.state


def test_run_default(cmd, test_cmd, ret_disabled):
    """
    ensure that python_shell defaults to False for cmd.run
    """
    ret = cmd.run(test_cmd)
    assert ret == ret_disabled


def test_shell_default(cmd, test_cmd, ret_enabled):
    """
    ensure that python_shell defaults to True for cmd.shell
    """
    ret = cmd.shell(test_cmd)
    assert ret == ret_enabled


def test_run_disabled(cmd, test_cmd, ret_disabled):
    """
    test python_shell disabled output for cmd.run
    """
    ret = cmd.run(test_cmd, python_shell=False)
    assert ret.strip() == ret_disabled


def test_shell_disabled(cmd, test_cmd, ret_disabled):
    """
    test python_shell disabled output for cmd.shell
    """
    ret = cmd.shell(test_cmd, python_shell=False)
    assert ret.strip() == ret_disabled


def test_run_enabled(cmd, test_cmd, ret_enabled):
    """
    test python_shell enabled output for cmd.run
    """
    ret = cmd.run(test_cmd, python_shell=True)
    assert ret.strip() == ret_enabled


def test_shell_enabled(cmd, test_cmd, ret_enabled):
    """
    test python_shell enabled output for cmd.shell
    """
    ret = cmd.shell(test_cmd, python_shell=True)
    assert ret.strip() == ret_enabled


@pytest.mark.slow_test
def test_template_run_default(state, state_tree, test_cmd, ret_disabled):
    """
    test python_shell disabled output for templates using run
    (python_shell=False is the default beginning with the 2017.7.0 release).
    """
    state_name = "template_run_default"
    state_file_name = state_name + ".sls"
    if salt.utils.platform.is_windows():
        test_cmd = f"'{test_cmd}'"
    else:
        test_cmd = f'"{test_cmd}"'
    state_file_contents = textwrap.dedent(
        f"""
        {{% set run_default= salt['cmd.run']({test_cmd}) %}}

        run_default:
          test.configurable_test_state:
            - name: '{{{{ run_default }}}}'
        """
    )

    # the result of running self.cmd not in a shell
    ret_disabled = ret_disabled.replace(os.linesep, " ")
    ret_key = f"test_|-run_default_|-{ret_disabled}_|-configurable_test_state"
    with pytest.helpers.temp_file(state_file_name, state_file_contents, state_tree):
        ret = state.sls(state_name)
        assert ret[ret_key]["name"] == ret_disabled


@pytest.mark.slow_test
def test_template_shell_default(state, state_tree, test_cmd, ret_enabled):
    """
    test python_shell enabled output for templates using shell
    (python_shell=True is the default beginning with the 2017.7.0 release).
    """
    state_name = "template_shell_enabled"
    state_file_name = state_name + ".sls"
    if salt.utils.platform.is_windows():
        test_cmd = f"'{test_cmd}'"
    else:
        test_cmd = f'"{test_cmd}"'
    state_file_contents = textwrap.dedent(
        f"""
        {{% set shell_default= salt['cmd.shell']({test_cmd}) %}}

        shell_default:
          test.configurable_test_state:
            - name: '{{{{ shell_default }}}}'
        """
    )

    # the result of running self.cmd not in a shell
    ret_enabled = ret_enabled.replace(os.linesep, " ")
    ret_key = f"test_|-shell_default_|-{ret_enabled}_|-configurable_test_state"
    with pytest.helpers.temp_file(state_file_name, state_file_contents, state_tree):
        ret = state.sls(state_name)
        assert ret[ret_key]["name"] == ret_enabled


@pytest.mark.slow_test
def test_template_run_disabled(state, state_tree, test_cmd, ret_disabled):
    """
    test python_shell disabled output for templates using run
    """
    state_name = "template_run_disabled"
    state_file_name = state_name + ".sls"
    if salt.utils.platform.is_windows():
        test_cmd = f"'{test_cmd}'"
    else:
        test_cmd = f'"{test_cmd}"'
    state_file_contents = textwrap.dedent(
        f"""
        {{% set run_disabled = salt['cmd.run']({test_cmd}, python_shell=False) %}}

        run_disabled:
          test.configurable_test_state:
            - name: '{{{{ run_disabled }}}}'
        """
    )

    # the result of running self.cmd not in a shell
    ret_disabled = ret_disabled.replace(os.linesep, " ")
    ret_key = f"test_|-run_disabled_|-{ret_disabled}_|-configurable_test_state"
    with pytest.helpers.temp_file(state_file_name, state_file_contents, state_tree):
        ret = state.sls(state_name)
        assert ret[ret_key]["name"] == ret_disabled


@pytest.mark.slow_test
def test_template_shell_disabled(state, state_tree, test_cmd, ret_disabled):
    """
    test python_shell disabled output for templates using shell
    """
    state_name = "template_shell_disabled"
    state_file_name = state_name + ".sls"
    if salt.utils.platform.is_windows():
        test_cmd = f"'{test_cmd}'"
    else:
        test_cmd = f'"{test_cmd}"'
    state_file_contents = textwrap.dedent(
        f"""
        {{% set shell_disabled = salt['cmd.shell']({test_cmd}, python_shell=False) %}}

        shell_disabled:
          test.configurable_test_state:
            - name: '{{{{ shell_disabled }}}}'
        """
    )

    # the result of running self.cmd not in a shell
    ret_disabled = ret_disabled.replace(os.linesep, " ")
    ret_key = f"test_|-shell_disabled_|-{ret_disabled}_|-configurable_test_state"
    with pytest.helpers.temp_file(state_file_name, state_file_contents, state_tree):
        ret = state.sls(state_name)
        assert ret[ret_key]["name"] == ret_disabled


@pytest.mark.slow_test
def test_template_run_enabled(state, state_tree, test_cmd, ret_enabled):
    """
    Test cmd.run works correctly when using a template.
    """
    state_name = "template_run_enabled"
    state_file_name = state_name + ".sls"
    if salt.utils.platform.is_windows():
        test_cmd = f"'{test_cmd}'"
    else:
        test_cmd = f'"{test_cmd}"'
    state_file_contents = textwrap.dedent(
        f"""
        {{% set run_enabled = salt['cmd.run']({test_cmd}, python_shell=True).strip() %}}

        run_enabled:
          test.configurable_test_state:
            - name: '{{{{ run_enabled }}}}'
        """
    )

    ret_enabled = ret_enabled.replace(os.linesep, " ")
    ret_key = f"test_|-run_enabled_|-{ret_enabled}_|-configurable_test_state"

    with pytest.helpers.temp_file(state_file_name, state_file_contents, state_tree):
        ret = state.sls(state_name)
        assert ret[ret_key]["name"] == ret_enabled


@pytest.mark.slow_test
def test_template_shell_enabled(state, state_tree, test_cmd, ret_enabled):
    """
    Test cmd.shell works correctly when using a template.

    Note: This test used to test that python_shell defaulted to True for templates
    in releases before 2017.7.0. The cmd.run --> cmd.shell aliasing was removed in
    2017.7.0. Templates should now be using cmd.shell.
    """
    state_name = "template_shell_enabled"
    state_file_name = state_name + ".sls"
    if salt.utils.platform.is_windows():
        test_cmd = f"'{test_cmd}'"
    else:
        test_cmd = f'"{test_cmd}"'
    state_file_contents = textwrap.dedent(
        f"""
        {{% set shell_enabled = salt['cmd.shell']({test_cmd}, python_shell=True).strip() %}}

        shell_enabled:
          test.configurable_test_state:
            - name: '{{{{ shell_enabled }}}}'
        """
    )

    ret_enabled = ret_enabled.replace(os.linesep, " ")
    ret_key = f"test_|-shell_enabled_|-{ret_enabled}_|-configurable_test_state"

    with pytest.helpers.temp_file(state_file_name, state_file_contents, state_tree):
        ret = state.sls(state_name)
        assert ret[ret_key]["name"] == ret_enabled



import pytest

import salt.modules.saltcheck as saltcheck
from tests.support.mock import MagicMock


@pytest.fixture()
def configure_loader_modules():
    return {saltcheck: {"__salt__": {"state.show_top": MagicMock()}}}


@pytest.mark.parametrize("saltenv", ["base", "dev", "howdy"])
def test__get_top_states_call_args(saltenv):
    saltcheck._get_top_states(saltenv=saltenv)
    saltcheck.__salt__["state.show_top"].assert_called_with(saltenv=saltenv)



"""
Management of Open vSwitch database records.

.. versionadded:: 3006.0
"""


def __virtual__():
    """
    Only make these states available if Open vSwitch module is available.
    """
    return "openvswitch.db_get" in __salt__


def managed(name, table, data, record=None):
    """
    Ensures that the specified columns of the named record have the specified
    values.

    Args:
        name : string
            name of the record
        table : string
            name of the table to which the record belongs.
        data : dict
            dictionary containing a mapping from column names to the desired
            values. Columns that exist, but are not specified in this
            dictionary are not touched.
        record : string
            name of the record (optional). Replaces name if specified.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    if record is None:
        record = name
    current_data = {
        column: __salt__["openvswitch.db_get"](table, record, column, True)
        for column in data
    }

    # Comment and change messages
    comment_changes = "Columns have been updated."
    comment_no_changes = "All columns are already up to date."
    comment_error = "Error while updating column {0}: {1}"

    # Dry run, test=true mode
    if __opts__["test"]:
        for column in data:
            if data[column] != current_data[column]:
                ret["changes"][column] = {
                    "old": current_data[column],
                    "new": data[column],
                }
        if ret["changes"]:
            ret["result"] = None
            ret["comment"] = comment_changes
        else:
            ret["result"] = True
            ret["comment"] = comment_no_changes
        return ret

    for column in data:
        if data[column] != current_data[column]:
            result = __salt__["openvswitch.db_set"](table, record, column, data[column])
            if result is not None:
                ret["comment"] = comment_error.format(column, result)
                ret["result"] = False
                return ret
            ret["changes"][column] = {"old": current_data[column], "new": data[column]}
    ret["result"] = True
    ret["comment"] = comment_no_changes
    return ret



"""
This package contains the loader modules for the salt streams system
"""

import copy
import logging
import re
import sys

import salt.utils.event
import salt.utils.minion

log = logging.getLogger(__name__)


class Beacon:
    """
    This class is used to evaluate and execute on the beacon system
    """

    def __init__(self, opts, functions):
        import salt.loader

        self.opts = opts
        self.functions = functions
        self.beacons = salt.loader.beacons(opts, functions)
        self.interval_map = dict()

    def process(self, config, grains):
        """
        Process the configured beacons

        The config must be a list and looks like this in yaml

        .. code_block:: yaml
            beacons:
              inotify:
                - files:
                    - /etc/fstab: {}
                    - /var/cache/foo: {}
        """
        ret = []
        b_config = copy.deepcopy(config)
        if "enabled" in b_config and not b_config["enabled"]:
            return
        for mod in config:
            if mod == "enabled":
                continue

            # Convert beacons that are lists to a dict to make processing easier
            current_beacon_config = None
            if isinstance(config[mod], list):
                current_beacon_config = {}
                list(map(current_beacon_config.update, config[mod]))
            elif isinstance(config[mod], dict):
                current_beacon_config = config[mod]

            if "enabled" in current_beacon_config:
                if not current_beacon_config["enabled"]:
                    log.trace("Beacon %s disabled", mod)
                    continue
                else:
                    # remove 'enabled' item before processing the beacon
                    if isinstance(config[mod], dict):
                        del config[mod]["enabled"]
                    else:
                        self._remove_list_item(config[mod], "enabled")

            log.trace("Beacon processing: %s", mod)
            beacon_name = None
            if self._determine_beacon_config(current_beacon_config, "beacon_module"):
                beacon_name = current_beacon_config["beacon_module"]
            else:
                beacon_name = mod

            # Run the validate function if it's available,
            # otherwise there is a warning about it being missing
            validate_str = f"{beacon_name}.validate"
            if validate_str in self.beacons:
                valid, vcomment = self.beacons[validate_str](b_config[mod])

                if not valid:
                    log.error(
                        "Beacon %s configuration invalid, not running.\n%s",
                        mod,
                        vcomment,
                    )
                    continue
            else:
                log.warning(
                    "No validate function found for %s, running basic beacon validation.",
                    mod,
                )
                if not isinstance(b_config[mod], list):
                    log.error("Configuration for beacon must be a list.")
                    continue

            b_config[mod].append({"_beacon_name": mod})
            fun_str = f"{beacon_name}.beacon"
            if fun_str in self.beacons:
                runonce = self._determine_beacon_config(
                    current_beacon_config, "run_once"
                )
                interval = self._determine_beacon_config(
                    current_beacon_config, "interval"
                )
                if interval:
                    b_config = self._trim_config(b_config, mod, "interval")
                    if not self._process_interval(mod, interval):
                        log.trace("Skipping beacon %s. Interval not reached.", mod)
                        continue
                if self._determine_beacon_config(
                    current_beacon_config, "disable_during_state_run"
                ):
                    log.trace(
                        "Evaluting if beacon %s should be skipped due to a state run.",
                        mod,
                    )
                    b_config = self._trim_config(
                        b_config, mod, "disable_during_state_run"
                    )
                    is_running = False
                    running_jobs = salt.utils.minion.running(self.opts)
                    for job in running_jobs:
                        if re.match("state.*", job["fun"]):
                            is_running = True
                    if is_running:
                        close_str = f"{beacon_name}.close"
                        if close_str in self.beacons:
                            log.info("Closing beacon %s. State run in progress.", mod)
                            self.beacons[close_str](b_config[mod])
                        else:
                            log.info("Skipping beacon %s. State run in progress.", mod)
                        continue
                # Update __grains__ on the beacon
                self.beacons[fun_str].__globals__["__grains__"] = grains

                # Fire the beacon!
                error = None
                try:
                    raw = self.beacons[fun_str](b_config[mod])
                except:  # pylint: disable=bare-except
                    error = f"{sys.exc_info()[1]}"
                    log.error("Unable to start %s beacon, %s", mod, error)
                    # send beacon error event
                    tag = "salt/beacon/{}/{}/".format(self.opts["id"], mod)
                    ret.append(
                        {
                            "tag": tag,
                            "error": error,
                            "data": {},
                            "beacon_name": beacon_name,
                        }
                    )
                if not error:
                    for data in raw:
                        tag = "salt/beacon/{}/{}/".format(self.opts["id"], mod)
                        if "tag" in data:
                            tag += data.pop("tag")
                        if "id" not in data:
                            data["id"] = self.opts["id"]
                        ret.append(
                            {"tag": tag, "data": data, "beacon_name": beacon_name}
                        )
                    if runonce:
                        self.disable_beacon(mod)
            else:
                log.warning("Unable to process beacon %s", mod)
        return ret

    def _trim_config(self, b_config, mod, key):
        """
        Take a beacon configuration and strip out the interval bits
        """
        if isinstance(b_config[mod], list):
            self._remove_list_item(b_config[mod], key)
        elif isinstance(b_config[mod], dict):
            b_config[mod].pop(key)
        return b_config

    def _determine_beacon_config(self, current_beacon_config, key):
        """
        Process a beacon configuration to determine its interval
        """

        interval = False
        if isinstance(current_beacon_config, dict):
            interval = current_beacon_config.get(key, False)

        return interval

    def _process_interval(self, mod, interval):
        """
        Process beacons with intervals
        Return True if a beacon should be run on this loop
        """
        log.trace("Processing interval %s for beacon mod %s", interval, mod)
        loop_interval = self.opts["loop_interval"]
        if mod in self.interval_map:
            log.trace("Processing interval in map")
            counter = self.interval_map[mod]
            log.trace("Interval counter: %s", counter)
            if counter * loop_interval >= interval:
                self.interval_map[mod] = 1
                return True
            else:
                self.interval_map[mod] += 1
        else:
            log.trace("Interval process inserting mod: %s", mod)
            self.interval_map[mod] = 1
        return False

    def _get_index(self, beacon_config, label):
        """
        Return the index of a labeled config item in the beacon config, -1 if the index is not found
        """

        indexes = [index for index, item in enumerate(beacon_config) if label in item]
        if not indexes:
            return -1
        else:
            return indexes[0]

    def _remove_list_item(self, beacon_config, label):
        """
        Remove an item from a beacon config list
        """

        index = self._get_index(beacon_config, label)
        del beacon_config[index]

    def _update_enabled(self, name, enabled_value):
        """
        Update whether an individual beacon is enabled
        """

        if isinstance(self.opts["beacons"][name], dict):
            # Backwards compatibility
            self.opts["beacons"][name]["enabled"] = enabled_value
        else:
            enabled_index = self._get_index(self.opts["beacons"][name], "enabled")
            if enabled_index >= 0:
                self.opts["beacons"][name][enabled_index]["enabled"] = enabled_value
            else:
                self.opts["beacons"][name].append({"enabled": enabled_value})

    def _get_beacons(self, include_opts=True, include_pillar=True):
        """
        Return the beacons data structure
        """
        beacons = {}
        if include_pillar:
            pillar_beacons = self.opts.get("pillar", {}).get("beacons", {})
            if not isinstance(pillar_beacons, dict):
                raise ValueError("Beacons must be of type dict.")
            beacons.update(pillar_beacons)
        if include_opts:
            opts_beacons = self.opts.get("beacons", {})
            if not isinstance(opts_beacons, dict):
                raise ValueError("Beacons must be of type dict.")
            beacons.update(opts_beacons)
        return beacons

    def list_beacons(self, include_pillar=True, include_opts=True):
        """
        List the beacon items

        include_pillar: Whether to include beacons that are
                        configured in pillar, default is True.

        include_opts:   Whether to include beacons that are
                        configured in opts, default is True.
        """
        beacons = self._get_beacons(
            include_pillar=include_pillar, include_opts=include_opts
        )

        # Fire the complete event back along with the list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {"complete": True, "beacons": beacons},
                tag="/salt/minion/minion_beacons_list_complete",
            )

        return True

    def list_available_beacons(self):
        """
        List the available beacons
        """
        _beacons = [
            "{}".format(_beacon.replace(".beacon", ""))
            for _beacon in self.beacons
            if ".beacon" in _beacon
        ]

        # Fire the complete event back along with the list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {"complete": True, "beacons": _beacons},
                tag="/salt/minion/minion_beacons_list_available_complete",
            )

        return True

    def validate_beacon(self, name, beacon_data):
        """
        Return available beacon functions
        """
        beacon_name = next(item.get("beacon_module", name) for item in beacon_data)

        validate_str = f"{beacon_name}.validate"
        # Run the validate function if it's available,
        # otherwise there is a warning about it being missing
        if validate_str in self.beacons:
            if "enabled" in beacon_data:
                del beacon_data["enabled"]
            valid, vcomment = self.beacons[validate_str](beacon_data)
        else:
            vcomment = (
                "Beacon {} does not have a validate"
                " function, skipping validation.".format(beacon_name)
            )
            valid = True

        # Fire the complete event back along with the list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {"complete": True, "vcomment": vcomment, "valid": valid},
                tag="/salt/minion/minion_beacon_validation_complete",
            )

        return True

    def add_beacon(self, name, beacon_data):
        """
        Add a beacon item
        """

        data = {}
        data[name] = beacon_data

        if name in self._get_beacons(include_opts=False):
            comment = (
                "Cannot update beacon item {}, "
                "because it is configured in pillar.".format(name)
            )
            complete = False
        else:
            if name in self.opts["beacons"]:
                comment = f"Updating settings for beacon item: {name}"
            else:
                comment = f"Added new beacon item: {name}"
            complete = True
            self.opts["beacons"].update(data)

        # Fire the complete event back along with updated list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {
                    "complete": complete,
                    "comment": comment,
                    "beacons": self.opts["beacons"],
                },
                tag="/salt/minion/minion_beacon_add_complete",
            )

        return True

    def modify_beacon(self, name, beacon_data):
        """
        Modify a beacon item
        """

        data = {}
        data[name] = beacon_data

        if name in self._get_beacons(include_opts=False):
            comment = f"Cannot modify beacon item {name}, it is configured in pillar."
            complete = False
        else:
            comment = f"Updating settings for beacon item: {name}"
            complete = True
            self.opts["beacons"].update(data)

        # Fire the complete event back along with updated list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {
                    "complete": complete,
                    "comment": comment,
                    "beacons": self.opts["beacons"],
                },
                tag="/salt/minion/minion_beacon_modify_complete",
            )
        return True

    def delete_beacon(self, name):
        """
        Delete a beacon item
        """

        if name in self._get_beacons(include_opts=False):
            comment = f"Cannot delete beacon item {name}, it is configured in pillar."
            complete = False
        else:
            if name in self.opts["beacons"]:
                del self.opts["beacons"][name]
                comment = f"Deleting beacon item: {name}"
            else:
                comment = f"Beacon item {name} not found."
            complete = True

        # Fire the complete event back along with updated list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {
                    "complete": complete,
                    "comment": comment,
                    "beacons": self.opts["beacons"],
                },
                tag="/salt/minion/minion_beacon_delete_complete",
            )

        return True

    def enable_beacons(self):
        """
        Enable beacons
        """

        self.opts["beacons"]["enabled"] = True

        # Fire the complete event back along with updated list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {"complete": True, "beacons": self.opts["beacons"]},
                tag="/salt/minion/minion_beacons_enabled_complete",
            )

        return True

    def disable_beacons(self):
        """
        Enable beacons
        """

        self.opts["beacons"]["enabled"] = False

        # Fire the complete event back along with updated list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {"complete": True, "beacons": self.opts["beacons"]},
                tag="/salt/minion/minion_beacons_disabled_complete",
            )

        return True

    def enable_beacon(self, name):
        """
        Enable a beacon
        """

        if name in self._get_beacons(include_opts=False):
            comment = f"Cannot enable beacon item {name}, it is configured in pillar."
            complete = False
        else:
            self._update_enabled(name, True)
            comment = f"Enabling beacon item {name}"
            complete = True

        # Fire the complete event back along with updated list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {
                    "complete": complete,
                    "comment": comment,
                    "beacons": self.opts["beacons"],
                },
                tag="/salt/minion/minion_beacon_enabled_complete",
            )

        return True

    def disable_beacon(self, name):
        """
        Disable a beacon
        """

        if name in self._get_beacons(include_opts=False):
            comment = (
                "Cannot disable beacon item {}, it is configured in pillar.".format(
                    name
                )
            )
            complete = False
        else:
            self._update_enabled(name, False)
            comment = f"Disabling beacon item {name}"
            complete = True

        # Fire the complete event back along with updated list of beacons
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {
                    "complete": complete,
                    "comment": comment,
                    "beacons": self.opts["beacons"],
                },
                tag="/salt/minion/minion_beacon_disabled_complete",
            )

        return True

    def reset(self):
        """
        Reset the beacons to defaults
        """
        self.opts["beacons"] = {}
        with salt.utils.event.get_event("minion", opts=self.opts) as evt:
            evt.fire_event(
                {
                    "complete": True,
                    "comment": "Beacons have been reset",
                    "beacons": self.opts["beacons"],
                },
                tag="/salt/minion/minion_beacon_reset_complete",
            )
        return True



"""
Fixtures for upgrade tests.
"""

import logging
import os
import pathlib
import shutil
import subprocess
import textwrap

import packaging.version
import pytest
import yaml
from pytestskipmarkers.utils import platform
from saltfactories.utils import random_string

import salt.config
import salt.utils.files
from tests.conftest import CODE_DIR, FIPS_TESTRUN
from tests.support.pkg import SaltMaster, SaltMasterWindows, SaltPkgInstall

log = logging.getLogger(__name__)


@pytest.fixture
def install_salt_systemd(request, salt_factories_root_dir):
    if platform.is_windows():
        conf_dir = "c:/salt/etc/salt"
    else:
        conf_dir = salt_factories_root_dir / "etc" / "salt"
    with SaltPkgInstall(
        conf_dir=conf_dir,
        pkg_system_service=request.config.getoption("--pkg-system-service"),
        upgrade=request.config.getoption("--upgrade"),
        downgrade=request.config.getoption("--downgrade"),
        no_uninstall=False,
        no_install=request.config.getoption("--no-install"),
        prev_version=request.config.getoption("--prev-version"),
        use_prev_version=request.config.getoption("--use-prev-version"),
    ) as fixture:
        # XXX Force un-install for now
        fixture.no_uninstall = False
        yield fixture


@pytest.fixture
def salt_factories_systemd(salt_factories, salt_factories_root_dir):
    salt_factories.root_dir = salt_factories_root_dir
    return salt_factories


@pytest.fixture
def master_systemd(salt_factories_systemd, install_salt_systemd, pkg_tests_account):
    """
    Start up a master
    """
    if platform.is_windows():
        state_tree = r"C:\salt\srv\salt"
        pillar_tree = r"C:\salt\srv\pillar"
    elif platform.is_darwin():
        state_tree = "/opt/srv/salt"
        pillar_tree = "/opt/srv/pillar"
    else:
        state_tree = "/srv/salt"
        pillar_tree = "/srv/pillar"

    start_timeout = None
    # Since the daemons are "packaged" with tiamat, the salt plugins provided
    # by salt-factories won't be discovered. Provide the required `*_dirs` on
    # the configuration so that they can still be used.
    config_defaults = {
        "engines_dirs": [
            str(salt_factories_systemd.get_salt_engines_path()),
        ],
        "log_handlers_dirs": [
            str(salt_factories_systemd.get_salt_log_handlers_path()),
        ],
    }
    if platform.is_darwin():
        config_defaults["enable_fqdns_grains"] = False
    config_overrides = {
        "timeout": 30,
        "file_roots": {
            "base": [
                state_tree,
            ]
        },
        "pillar_roots": {
            "base": [
                pillar_tree,
            ]
        },
        "rest_cherrypy": {
            "port": 8000,
            "disable_ssl": True,
        },
        "netapi_enable_clients": ["local"],
        "external_auth": {
            "auto": {
                pkg_tests_account.username: [
                    ".*",
                ],
            },
        },
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
        "open_mode": True,
    }
    salt_user_in_config_file = False
    master_config = install_salt_systemd.config_path / "master"
    if master_config.exists() and master_config.stat().st_size:
        with salt.utils.files.fopen(master_config) as fp:
            data = yaml.safe_load(fp)
            if data is None:
                # File exists but is mostly commented out
                data = {}
            user_in_config_file = data.get("user")
            if user_in_config_file and user_in_config_file == "salt":
                salt_user_in_config_file = True
                # We are testing a different user, so we need to test the system
                # configs, or else permissions will not be correct.
                config_overrides["user"] = user_in_config_file
                config_overrides["log_file"] = salt.config.DEFAULT_MASTER_OPTS.get(
                    "log_file"
                )
                config_overrides["root_dir"] = salt.config.DEFAULT_MASTER_OPTS.get(
                    "root_dir"
                )
                config_overrides["key_logfile"] = salt.config.DEFAULT_MASTER_OPTS.get(
                    "key_logfile"
                )
                config_overrides["pki_dir"] = salt.config.DEFAULT_MASTER_OPTS.get(
                    "pki_dir"
                )
                config_overrides["api_logfile"] = salt.config.DEFAULT_API_OPTS.get(
                    "api_logfile"
                )
                config_overrides["api_pidfile"] = salt.config.DEFAULT_API_OPTS.get(
                    "api_pidfile"
                )
                # verify files were set with correct owner/group
                verify_files = [
                    pathlib.Path("/etc", "salt", "pki", "master"),
                    pathlib.Path("/etc", "salt", "master.d"),
                    pathlib.Path("/var", "cache", "salt", "master"),
                ]
                for _file in verify_files:
                    if _file.owner() != "salt":
                        log.warning(
                            "The owner of '%s' is '%s' when it should be 'salt'",
                            _file,
                            _file.owner(),
                        )
                    if _file.group() != "salt":
                        log.warning(
                            "The group of '%s' is '%s' when it should be 'salt'",
                            _file,
                            _file.group(),
                        )

    master_script = False
    if platform.is_windows():
        master_script = True

    if master_script:
        salt_factories_systemd.system_service = False
        salt_factories_systemd.generate_scripts = True
        scripts_dir = salt_factories_systemd.root_dir / "Scripts"
        scripts_dir.mkdir(exist_ok=True)
        salt_factories_systemd.scripts_dir = scripts_dir
        python_executable = install_salt_systemd.install_dir / "Scripts" / "python.exe"
        salt_factories_systemd.python_executable = python_executable
        factory = salt_factories_systemd.salt_master_daemon(
            "pkg-test-master",
            defaults=config_defaults,
            overrides=config_overrides,
            factory_class=SaltMasterWindows,
            salt_pkg_install=install_salt_systemd,
        )
        salt_factories_systemd.system_service = True
    else:
        factory = salt_factories_systemd.salt_master_daemon(
            "pkg-test-master",
            defaults=config_defaults,
            overrides=config_overrides,
            factory_class=SaltMaster,
            salt_pkg_install=install_salt_systemd,
        )
    factory.after_terminate(pytest.helpers.remove_stale_master_key, factory)
    if salt_user_in_config_file:
        # Salt factories calls salt.utils.verify.verify_env
        # which sets root perms on /etc/salt/pki/master since we are running
        # the test suite as root, but we want to run Salt master as salt
        # We ensure those permissions where set by the package earlier
        subprocess.run(
            [
                "chown",
                "-R",
                "salt:salt",
                str(pathlib.Path("/etc", "salt", "pki", "master")),
            ],
            check=True,
        )

        if not platform.is_windows() and not platform.is_darwin():
            # The engines_dirs is created in .nox path. We need to set correct perms
            # for the user running the Salt Master
            check_paths = [state_tree, pillar_tree, CODE_DIR / ".nox"]
            for path in check_paths:
                if os.path.exists(path) is False:
                    continue
                subprocess.run(["chown", "-R", "salt:salt", str(path)], check=False)

    with factory.started(start_timeout=start_timeout):
        yield factory


@pytest.fixture
def minion_systemd(salt_factories_systemd, master_systemd, install_salt_systemd):
    """
    Start up a minion
    """
    start_timeout = None
    minion_id = random_string("minion-")
    # Since the daemons are "packaged" with tiamat, the salt plugins provided
    # by salt-factories won't be discovered. Provide the required `*_dirs` on
    # the configuration so that they can still be used.
    config_defaults = {
        "engines_dirs": master_systemd.config["engines_dirs"].copy(),
        "log_handlers_dirs": master_systemd.config["log_handlers_dirs"].copy(),
    }
    if platform.is_darwin():
        config_defaults["enable_fqdns_grains"] = False
    config_overrides = {
        "id": minion_id,
        "file_roots": master_systemd.config["file_roots"].copy(),
        "pillar_roots": master_systemd.config["pillar_roots"].copy(),
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    if platform.is_windows():
        config_overrides["winrepo_dir"] = (
            rf"{salt_factories_systemd.root_dir}\srv\salt\win\repo"
        )
        config_overrides["winrepo_dir_ng"] = (
            rf"{salt_factories_systemd.root_dir}\srv\salt\win\repo_ng"
        )
        config_overrides["winrepo_source_dir"] = r"salt://win/repo_ng"

    factory = master_systemd.salt_minion_daemon(
        minion_id,
        overrides=config_overrides,
        defaults=config_defaults,
    )

    # Salt factories calls salt.utils.verify.verify_env
    # which sets root perms on /srv/salt and /srv/pillar since we are running
    # the test suite as root, but we want to run Salt master as salt
    if not platform.is_windows() and not platform.is_darwin():
        import pwd

        try:
            pwd.getpwnam("salt")
        except KeyError:
            # The salt user does not exist
            pass
        else:
            state_tree = "/srv/salt"
            pillar_tree = "/srv/pillar"
            check_paths = [state_tree, pillar_tree, CODE_DIR / ".nox"]
            for path in check_paths:
                if os.path.exists(path) is False:
                    continue
                subprocess.run(["chown", "-R", "salt:salt", str(path)], check=False)

    if platform.is_windows():
        minion_pki = pathlib.Path("c:/salt/etc/salt/pki")
        if minion_pki.exists():
            salt.utils.files.rm_rf(minion_pki)

        # Work around missing WMIC until 3008.10 has been released. Not sure
        # why this doesn't work anymore on the master branch when it was enough
        # on 3006.x and 3007.x. We had to add similar logic in
        # tests/support/pkg.py to fix the upgrade/downgrade tests on master.
        grainsdir = pathlib.Path("c:/salt/etc/grains")
        grainsdir.mkdir(exist_ok=True)
        shutil.copy(r"salt\grains\disks.py", grainsdir)

        grainsdir = pathlib.Path(
            r"C:\Program Files\Salt Project\Salt\Lib\site-packages\salt\grains"
        )
        shutil.copy(r"salt\grains\disks.py", grainsdir)

    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, master_systemd, factory.id
    )
    with factory.started(start_timeout=start_timeout):
        yield factory


@pytest.fixture
def call_cli(minion_systemd):
    return minion_systemd.salt_call_cli()


@pytest.fixture
def salt_systemd_overrides():
    """
    Fixture to create systemd overrides for salt-api, salt-minion, and
    salt-master services.

    This is required because the pytest-salt-factories engine does not
    stop cleanly if you only kill the process. This leaves the systemd
    service in a failed state.
    """

    systemd_dir = pathlib.Path("/etc/systemd/system")
    conf_name = "override.conf"
    contents = textwrap.dedent(
        """
        [Service]
        KillMode=control-group
        TimeoutStopSec=10
        SuccessExitStatus=SIGKILL
        """
    )
    assert not (systemd_dir / "salt-api.service.d" / conf_name).exists()

    with pytest.helpers.temp_file(
        name=conf_name, directory=systemd_dir / "salt-api.service.d", contents=contents
    ), pytest.helpers.temp_file(
        name=conf_name,
        directory=systemd_dir / "salt-minion.service.d",
        contents=contents,
    ), pytest.helpers.temp_file(
        name=conf_name,
        directory=systemd_dir / "salt-master.service.d",
        contents=contents,
    ):
        yield
    assert not (systemd_dir / "salt-api.service.d" / conf_name).exists()


@pytest.fixture
def salt_systemd_setup(
    call_cli, install_salt_systemd, salt_systemd_overrides, debian_disable_policy_rcd
):
    """
    Fixture install previous version and set systemd for salt packages
    to enabled and active

    This fixture is function scoped, so it will be run for each test
    """

    upgrade_version = packaging.version.parse(install_salt_systemd.artifact_version)
    test_list = ["salt-api", "salt-minion", "salt-master"]

    # We should have a previous version installed, but if not then use install_previous
    ret = call_cli.run("--local", "test.version")
    assert ret.returncode == 0
    installed_minion_version = packaging.version.parse(ret.data)
    if installed_minion_version >= upgrade_version:
        # Install previous version, downgrading if necessary
        install_salt_systemd.install_previous(downgrade=True)

    # Verify that the previous version is installed
    ret = call_cli.run("--local", "test.version")
    assert ret.returncode == 0
    installed_minion_version = packaging.version.parse(ret.data)
    assert installed_minion_version < upgrade_version
    previous_version = installed_minion_version

    # Ensure known state for systemd services - enabled
    for test_item in test_list:
        test_cmd = f"systemctl enable {test_item}"
        ret = call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0

    # Run tests
    yield

    # Verify that the new version is installed after the test
    ret = call_cli.run("--local", "test.version")
    assert ret.returncode == 0
    installed_minion_version = packaging.version.parse(ret.data)
    assert installed_minion_version == upgrade_version

    # Reset systemd services to their preset states
    for test_item in test_list:
        test_cmd = f"systemctl preset {test_item}"
        ret = call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0

    # Install previous version, downgrading if necessary
    install_salt_systemd.install_previous(downgrade=True)

    # Ensure services are started on debian/ubuntu
    if install_salt_systemd.distro_name in ["debian", "ubuntu"]:
        install_salt_systemd.restart_services()

    # For debian/ubuntu, ensure pinning file is for major version of previous
    # version, not minor
    if install_salt_systemd.distro_name in ["debian", "ubuntu"]:
        pref_file = pathlib.Path("/etc", "apt", "preferences.d", "salt-pin-1001")
        pref_file.parent.mkdir(exist_ok=True)
        pin = f"{previous_version.major}.*"
        with salt.utils.files.fopen(pref_file, "w") as fp:
            fp.write(f"Package: salt-*\n" f"Pin: version {pin}\n" f"Pin-Priority: 1001")


@pytest.fixture
def salt_systemd_mask_services(call_cli):
    """
    Fixture to mask systemd services for salt-api, salt-minion, and
    salt-master services.

    This is required to test the preservation of masked state during upgrades.
    """

    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl mask {test_item}"
        ret = call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0

    yield

    # Cleanup: unmask the services after the test
    for test_item in test_list:
        test_cmd = f"systemctl unmask {test_item}"
        ret = call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0



"""
Tests for salt.loader.context
"""

import copy

import salt.loader.context
import salt.loader.lazy


def test_named_loader_context():
    loader_context = salt.loader.context.LoaderContext()
    named_context = salt.loader.context.NamedLoaderContext("__test__", loader_context)
    test_dunder = {"foo": "bar"}
    lazy_loader = salt.loader.lazy.LazyLoader(["/foo"], pack={"__test__": test_dunder})
    assert named_context.loader() is None
    token = salt.loader.context.loader_ctxvar.set(lazy_loader)
    try:
        assert named_context.loader() == lazy_loader
        # The loader's value is the same object as test_dunder
        assert named_context.value() is test_dunder
        assert named_context["foo"] == "bar"
    finally:
        salt.loader.context.loader_ctxvar.reset(token)


def test_named_loader_default():
    loader_context = salt.loader.context.LoaderContext()
    default = {"foo": "bar"}
    named_context = salt.loader.context.NamedLoaderContext(
        "__test__", loader_context, default=default
    )
    assert named_context.loader() is None
    # The loader's value is the same object as default
    assert named_context.value() is default
    assert named_context["foo"] == "bar"


def test_named_loader_context_deepcopy():
    loader_context = salt.loader.context.LoaderContext()
    default_data = {"foo": "bar"}
    named_context = salt.loader.context.NamedLoaderContext(
        "__test__", loader_context, default_data
    )
    coppied = copy.deepcopy(named_context)
    assert coppied.name == named_context.name
    assert id(coppied.loader_context) == id(named_context.loader_context)
    assert id(coppied.default) != id(named_context.default)


def test_named_loader_context_opts():
    loader_context = salt.loader.context.LoaderContext()
    opts = loader_context.named_context("__opts__")
    loader = salt.loader.lazy.LazyLoader(["/foo"], opts={"foo": "bar"})
    with salt.loader.context.loader_context(loader):
        assert "foo" in opts
        assert opts["foo"] == "bar"



"""
Unit tests for the docker state
"""

import pytest

import salt.modules.dockermod as docker_mod
import salt.states.docker_image as docker_state
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        docker_mod: {"__context__": {"docker.docker_version": ""}},
        docker_state: {"__opts__": {"test": False}},
    }


def test_present_already_local():
    """
    According following sls,

    .. code-block:: yaml

        image:latest:
          docker_image.present:
            - force: true

    if ``image:latest`` is already downloaded locally the state
    should not report changes.
    """
    docker_inspect_image = MagicMock(return_value={"Id": "abcdefghijkl"})
    docker_pull = MagicMock(
        return_value={
            "Layers": {"Already_Pulled": ["abcdefghijkl"], "Pulled": []},
            "Status": "Image is up to date for image:latest",
            "Time_Elapsed": 1.1,
        }
    )
    docker_list_tags = MagicMock(return_value=["image:latest"])
    docker_resolve_tag = MagicMock(return_value="image:latest")
    __salt__ = {
        "docker.list_tags": docker_list_tags,
        "docker.pull": docker_pull,
        "docker.inspect_image": docker_inspect_image,
        "docker.resolve_tag": docker_resolve_tag,
    }
    with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
        ret = docker_state.present("image:latest", force=True)
        assert ret == {
            "changes": {},
            "result": True,
            "comment": "Image 'image:latest' was pulled, but there were no changes",
            "name": "image:latest",
        }


def test_present_and_force():
    """
    According following sls,

    .. code-block:: yaml

        image:latest:
          docker_image.present:
            - force: true

    if ``image:latest`` is not downloaded and force is true
    should pull a new image successfully.
    """
    docker_inspect_image = MagicMock(return_value={"Id": "1234567890ab"})
    docker_pull = MagicMock(
        return_value={
            "Layers": {"Pulled": ["abcdefghijkl"]},
            "Status": "Image 'image:latest' was pulled",
            "Time_Elapsed": 1.1,
        }
    )
    docker_list_tags = MagicMock(side_effect=[[], ["image:latest"]])
    docker_resolve_tag = MagicMock(return_value="image:latest")
    __salt__ = {
        "docker.list_tags": docker_list_tags,
        "docker.pull": docker_pull,
        "docker.inspect_image": docker_inspect_image,
        "docker.resolve_tag": docker_resolve_tag,
    }
    with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
        ret = docker_state.present("image:latest", force=True)
    assert ret == {
        "changes": {
            "Layers": {"Pulled": ["abcdefghijkl"]},
            "Status": "Image 'image:latest' was pulled",
            "Time_Elapsed": 1.1,
        },
        "result": True,
        "comment": "Image 'image:latest' was pulled",
        "name": "image:latest",
    }
